
-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: C_EncounterJournal, C_RaidLocks, C_Scenario, C_SocialQueue, ChatEdit_ActivateChat, ChatEdit_ChooseBoxForSend, ChatEdit_GetActiveWindow, ChatFrame_GetMessageEventFilters, ChatFrame_OpenChat, ChatFrame_SendTell, CreateColor, DEFAULT_CHAT_FRAME, EJ_GetEncounterInfoByIndex, EJ_SelectInstance, GetInstanceInfo, GetNumGroupMembers, GetNumSubgroupMembers, GetRelativeDifficultyColor, GetUnitName, GRAY_FONT_COLOR, IsAltKeyDown, IsControlKeyDown, IsModifierKeyDown, IsShiftKeyDown, next, select, SettingsPanel, SocialQueueUtil_HasRelationshipWithLeader, table, tonumber, type, UnitEffectiveLevel, UnitInParty, UnitInRaid, UnitIsFriend, UnitIsPlayer, UnitLevel

local C_Map = C_Map
local CHALLENGE_MODE_ITEM_POWER_LEVEL = CHALLENGE_MODE_ITEM_POWER_LEVEL
local COPPER_AMOUNT_SYMBOL = COPPER_AMOUNT_SYMBOL
local COPPER_PER_GOLD = COPPER_PER_GOLD
local COPPER_PER_SILVER = COPPER_PER_SILVER
local DUNGEON_NAME_WITH_DIFFICULTY = DUNGEON_NAME_WITH_DIFFICULTY
local EJ_GetInstanceForMap = EJ_GetInstanceForMap
local floor = floor
local format = format
local GetNormalizedRealmName = GetNormalizedRealmName
local GOLD_AMOUNT_SYMBOL = GOLD_AMOUNT_SYMBOL
local SILVER_AMOUNT_SYMBOL = SILVER_AMOUNT_SYMBOL
local string = string
local UNKNOWN = UNKNOWN

-------------------------------------------------------------------------------
-- Quick Join
-------------------------------------------------------------------------------

-- Show a Quick Join tooltip.
function A:ShowGroupTooltip(tooltip, name, groupGUID)
	local canJoin, _, _, _, _, _, _, leaderGUID = C_SocialQueue.GetGroupInfo(groupGUID)
	if not canJoin then return end
	local queues = C_SocialQueue.GetGroupQueues(groupGUID)
	if not queues then return end
	local knowsLeader = SocialQueueUtil_HasRelationshipWithLeader(leaderGUID)
	SocialQueueUtil_SetTooltip(tooltip, name, queues, canJoin, knowsLeader)
end 

-------------------------------------------------------------------------------
-- Click Handler
-------------------------------------------------------------------------------

function A:GetModifierKey()
	return (not IsModifierKeyDown() and C.Enum.ModifierKey.None)
		or (IsAltKeyDown() and C.Enum.ModifierKey.Alt)
		or (IsControlKeyDown() and C.Enum.ModifierKey.Control)
		or (IsShiftKeyDown() and C.Enum.ModifierKey.Shift)
		or C.Enum.ModifierKey.Unknown
end

function A:GetMouseButton(button)
	return (button == "LeftButton" and C.Enum.MouseButton.Left)
		or (button == "MiddleButton" and C.Enum.MouseButton.Middle)
		or (button == "RightButton" and C.Enum.MouseButton.Right)
		or C.Enum.MouseButton.Unknown
end

function A:HandleClick(table, button, ...)
	local modifierKey = self:GetModifierKey()
	local mouseButton = self:GetMouseButton(button)

	for _, entry in next, table do
		if (type(entry.mouseButton) == "nil") or (entry.mouseButton == mouseButton) or (entry.mouseButton == C.Enum.MouseButton.Any) then
			if (type(entry.modifierKey) == "nil") or (entry.modifierKey == modifierKey) or (entry.modifierKey == C.Enum.ModifierKey.Any) then
				if (type(entry.test) == "nil") or (entry.test(...) == true) then
					return entry.callback(...)
				end
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Utility
-------------------------------------------------------------------------------

function A.CloseEditBox()
	local editBox = ChatEdit_ChooseBoxForSend()
	editBox:ClearFocus()
end

function A.FilterTable(t, predicateFn)
	local filteredTable = {}

	for _, v in next, t do
		if predicateFn(v) then
			table.insert(filteredTable, v)
		end
	end

	return filteredTable
end

function A.FlavourStateText(stateText)
	local clutter = C.CLUTTER_KEY_STATE_TEXT
	local noclutter = clutter:gsub("mod:", "nomod:")

	stateText = stateText:gsub("noclutter", noclutter)
	stateText = stateText:gsub("clutter", clutter)

	return stateText
end

-- Convert 1.234567 into 1.23.
function A.FormatDecimal(value)
	return floor(value * 100) / 100
end

-- Returns a formatted string with the instance name, difficulty, and Keystone level (if applicable).
function A.FormatInstanceName(name, difficultyName, keystoneLevel)
	return DUNGEON_NAME_WITH_DIFFICULTY:format(name, keystoneLevel and keystoneLevel ~= 0 and CHALLENGE_MODE_ITEM_POWER_LEVEL:format(keystoneLevel) or difficultyName or UNKNOWN)
end

-- Returns "Name" if the realm is the player's realm, or "Name-Realm".
function A.FormatUnit(name, realm, alwaysReturnRealm)
	if not C.PLAYER_REALM_NAME then
		C.PLAYER_REALM_NAME = GetNormalizedRealmName()
	end

	if name then
		if name:match("-") then
			name, realm = string.split("-", name)
		end

		if not realm or realm == "" then
			realm = C.PLAYER_REALM_NAME
		end

		if not alwaysReturnRealm and realm == C.PLAYER_REALM_NAME then
			return name
		end

		return name .. "-" .. realm
	end
end

function A.FormatMoney(money)
	local gold, symbol

	if money >= C.COPPER_PER_MILLION_GOLD then
		gold = A.FormatDecimal(money / C.COPPER_PER_MILLION_GOLD)
		symbol = "m"
	elseif money >= C.COPPER_PER_THOUSAND_GOLD then
		gold = floor(money / C.COPPER_PER_THOUSAND_GOLD)
		symbol = "k"
	elseif money >= COPPER_PER_GOLD then
		gold = floor(money / COPPER_PER_GOLD)
		symbol = GOLD_AMOUNT_SYMBOL
	elseif money >= COPPER_PER_SILVER then
		gold = floor(money / COPPER_PER_SILVER)
		symbol = SILVER_AMOUNT_SYMBOL
	else
		gold = money
		symbol = COPPER_AMOUNT_SYMBOL
	end

	local formatted = string.format("%s|cff%s%s|r", gold, symbol == "c" and "eda55f" or symbol == "s" and "c7c7cf" or "ffd700", symbol)

	return formatted, gold, symbol
end

function A.GetJournalBosses(journalInstanceID)
	if (not journalInstanceID) then return end
	local bosses = {}
	local difficultyID = select(3, GetInstanceInfo())
	local isInScenario = C_Scenario.IsInScenario()

	EJ_SelectInstance(journalInstanceID)

	for index = 1, 20 do
		local name, _, journalEncounterID, _, _, _, encounterID, instanceID = EJ_GetEncounterInfoByIndex(index, journalInstanceID)

		if name then
			local isDefeated = C_EncounterJournal.IsEncounterComplete(journalEncounterID)
			local isLocked = not isInScenario and C_RaidLocks.IsEncounterComplete(instanceID, encounterID, difficultyID)
			local status = (isInScenario and isDefeated and C.Enum.BossStatus.Dead)
				or (isInScenario and C.Enum.BossStatus.Alive)
				or (isDefeated and isLocked and C.Enum.BossStatus.Dead)
				or (isDefeated and C.Enum.BossStatus.DeadUnsaved)
				or (isLocked and C.Enum.BossStatus.AliveSaved)
				or C.Enum.BossStatus.Alive

			table.insert(bosses, {name = name, status = status})
		end
	end

	return bosses
end

function A.GetJournalInstanceID()
	local mapID = C_Map.GetBestMapForUnit("player")
	local journalInstanceID = mapID and EJ_GetInstanceForMap(mapID) or 0
	if (journalInstanceID ~= 0) then return journalInstanceID end
end

-- Get a frame's closest screen corner.
function A.GetScreenCorner(frame)
	local point, _, _, offsetX, offsetY = frame:GetPoint(1)

	if not (point == "BOTTOMRIGHT" or point == "BOTTOMLEFT" or point == "TOPRIGHT" or point == "TOPLEFT") then
		if point == "TOP" or point == "BOTTOM" then
			point = offsetX > 0 and point .. "RIGHT" or point .. "LEFT"
		elseif point =="LEFT" or point == "RIGHT" then
			point = offsetY > 0 and "TOP" .. point or "BOTTOM" .. point
		else
			local topBottom = offsetY > 0 and "TOP" or "BOTTOM"
			local rightLeft = offsetX > 0 and "RIGHT" or "LEFT"

			point = topBottom .. rightLeft;
		end
	end

	return point
end

function A.GetSettingsCategoryID(name)
	for _, category in next, SettingsPanel:GetCategoryList():GetAllCategories() do
		if category.name == name then
			return category.ID
		end
	end
end

-- Get the relative difficulty color of a level.
function A.GetDifficultyColor(difficultyLevel, isScaling)
	if tonumber(difficultyLevel) then
		local playerLevel = isScaling and UnitEffectiveLevel("player") or UnitLevel("player")
		local color = GetRelativeDifficultyColor(playerLevel, difficultyLevel)

		return CreateColor(color.r, color.g, color.b)
	end

	return GRAY_FONT_COLOR
end

-- Wrap text in a color code.
function A.Highlight(text, color)
	return (color or C.HIGHLIGHT_UTILITY_COLOR):WrapTextInColorCode(text or "")
end

-- Check whether an event is being filtered.
function A.IsFiltered(event, ...)
	local filters = ChatFrame_GetMessageEventFilters(event)

	for _, filterFunc in next, filters do
		if filterFunc(DEFAULT_CHAT_FRAME, event, ...) then
			return true
		end
	end

	return false
end

-- Open the edit box.
function A.OpenEditBox()
	ChatEdit_ActivateChat(ChatEdit_ChooseBoxForSend())
end

-- Print a message to the default chat frame.
function A.Print(message)
	DEFAULT_CHAT_FRAME:AddMessage(message, 1, 1, 1)
end

-- Print a formatted message to the default chat frame.
function A.PrintF(message, ...)
	A.Print(format(message, ...))
end

-- Set the edit box's text.
function A.SetEditBoxText(text)
	local editBox = ChatEdit_ChooseBoxForSend()

	if editBox ~= ChatEdit_GetActiveWindow() then
		ChatFrame_OpenChat(text)
	else
		editBox:SetFocus()
		editBox:SetText(text)
		editBox:HighlightText()
	end
end

-- Send a tell to the player's target.
function A.TellTarget()
	if UnitIsPlayer("target") and UnitIsFriend("player", "target") then
		ChatFrame_SendTell(GetUnitName("target", true))
	end
end

-- Check whether a unit is in the player's raid or party.
function A.UnitInGroup(unit)
	if unit then
		return (GetNumGroupMembers() and UnitInRaid(unit)) or (GetNumSubgroupMembers() and UnitInParty(unit))
	end
end
